import json
from typing import Any, Optional, Callable, Dict

from .exceptions import UnsupportedMediaType

MediaHandler = Callable[[Any], str]


def handle_json(value: dict) -> str:
    return json.dumps(value)


def handle_text(value: Any) -> str:
    return str(value)


def get_default_handlers() -> dict:
    return {
        Media.JSON: handle_json,
        Media.PLAIN_TEXT: handle_text,
        Media.HTML: handle_text,
    }


class Media:
    """Registry of media handlers."""

    JSON = 'application/json'
    PLAIN_TEXT = 'text/plain'
    HTML = 'text/html'

    def __init__(self, media_type: str,
                 handlers: Dict[str, MediaHandler] = None):
        """Create a media registry.

        Parameters
        -----------
        media_type : str
            The default media type that will be used when serializing values.
        handlers : dict of str -> MediaHandler, optional
            A mapping of media type to an (Any) -> str callable.
            Defaults to built-in media handlers.
        """
        if handlers is None:
            handlers = get_default_handlers()
        self.handlers = handlers
        self.type = media_type

    def serialize(self, value: Any, media_type: Optional[str] = None):
        """Serialize a value using the given media type.

        Parameters
        ----------
        value : any
        media_type : str, optional
            The media type of the given value. Determines which media handler
            is used.
            Defaults to the media registry's media type.
        """
        if media_type is None:
            media_type = self.type
        handler = self.handlers[media_type]
        return handler(value)

    @property
    def type(self) -> str:
        """Return the default media type."""
        return self._default_type

    @type.setter
    def type(self, media_type: str):
        """Set the default media type.

        Parameters
        ----------
        media_type : str

        Raises
        ------
        UnsupportedMediaType :
            If no handler exists for the given media type.
        """
        if media_type not in self.handlers:
            raise UnsupportedMediaType(media_type,
                                       available=list(self.handlers))
        self._default_type = media_type



from collections import defaultdict
from functools import wraps
from http import HTTPStatus
from typing import Optional, List, Union, Callable, Dict

from parse import parse

from .compat import call_async
from .exceptions import HTTPError
from .hooks import HookFunction, BEFORE, AFTER, empty_hook
from .view import View, create_callable_view


class Route:
    """Represents a route to a view.

    Formatted string syntax is used for route patterns.
    """

    def __init__(self, pattern: str, view: View, methods: List[str], name: str):
        self._pattern = pattern

        self._view = create_callable_view(view=view)
        self._methods = methods
        self._name = name

        self.hooks: Dict[str, HookFunction] = defaultdict(lambda: empty_hook)

    def url(self, **kwargs) -> str:
        """Return full path for the given route parameters."""
        return self._pattern.format(**kwargs)

    def match(self, path: str) -> Optional[dict]:
        """Return whether the route matches the given path.

        Examples
        -------
        >>> route = Route('/{age:d}', lambda req, res: None)
        >>> route.match('/42')
        {'age': 42}
        >>> route.match('/john')
        None
        """
        result = parse(self._pattern, path)
        if result is not None:
            return result.named
        return None

    @classmethod
    def before_hook(cls, hook_function: HookFunction, *args, **kwargs):
        return cls._add_hook(BEFORE, hook_function, *args, **kwargs)

    @classmethod
    def after_hook(cls, hook_function: HookFunction, *args, **kwargs):
        return cls._add_hook(AFTER, hook_function, *args, **kwargs)

    @staticmethod
    def _add_hook(hook: str, hook_function: HookFunction, *args, **kwargs):
        def decorator(hookable: Union[Route, Callable]):
            """Bind the hook function to the given hookable object.

            Support for decorating a route or a class method enables
            using hooks in the following contexts:
            - On a function-based view (before @api.route()).
            - On top of a class-based view (before @api.route()).
            - On a class-based view method.

            Parameters
            ----------
            hookable : Route or (unbound) class method
            """
            nonlocal hook_function
            full_hook_function = hook_function

            async def hook_function(req, res, params):
                return await call_async(
                    full_hook_function, req, res, params, *args, **kwargs
                )

            if isinstance(hookable, Route):
                route = hookable
                route.hooks[hook] = hook_function
                return route
            else:
                view: Callable = hookable

                @wraps(view)
                async def with_hook(self, req, res, **kw):
                    if hook == BEFORE:
                        await hook_function(req, res, kw)
                    await call_async(view, self, req, res, **kw)
                    if hook == AFTER:
                        await hook_function(req, res, kw)

                return with_hook

        return decorator

    def raise_for_method(self, request):
        if request.method not in self._methods:
            raise HTTPError(status=HTTPStatus.METHOD_NOT_ALLOWED)

    async def __call__(self, request, response, **kwargs) -> None:
        view = self._view
        await call_async(self.hooks[BEFORE], request, response, kwargs)
        await view(request, response, **kwargs)
        await call_async(self.hooks[AFTER], request, response, kwargs)



from typing import Callable

from .request import Request
from .response import Response
from .view import View

HookFunction = Callable[[Request, Response, View, dict], None]

BEFORE = 'before'
AFTER = 'after'


def empty_hook(req, res, params):
    pass



import asyncio
import inspect
from typing import Callable, Union, Coroutine

from .compat import call_async
from .constants import ALL_HTTP_METHODS
from .request import Request
from .response import Response


class ClassBasedView:
    """Class-based view interface."""

    def get(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def post(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def put(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def patch(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def delete(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError


# Types
CallableView = Callable[[Request, Response, dict], Coroutine]
View = Union[CallableView, ClassBasedView]


def create_callable_view(view: View) -> CallableView:
    """Create callable view from function (sync/async) or class based view."""
    if asyncio.iscoroutinefunction(view):
        return view
    elif inspect.isfunction(view):
        async def callable_view(req, res, **kwargs):
            await call_async(view, req, res, sync=True, **kwargs)

        return callable_view
    else:
        return _from_class_instance(view)


def _from_class_instance(view: ClassBasedView):
    def _find_for_method(method: str):
        try:
            return getattr(view, 'handle')
        except AttributeError:
            return getattr(view, method.lower())

    async def callable_view(req, res, **kwargs):
        view_ = _find_for_method(req.method)
        await call_async(view_, req, res, **kwargs)

    return callable_view


def get_view_name(view: View, base: ClassBasedView = None) -> str:
    def _get_name(obj):
        return getattr(obj, '__name__', obj.__class__.__name__)

    return '.'.join(_get_name(part) for part in (base, view) if part)


def get_declared_method_views(view: ClassBasedView):
    for method in ('handle', *map(str.lower, ALL_HTTP_METHODS)):
        if hasattr(view, method):
            yield method, getattr(view, method)



import asyncio
from typing import Callable, Coroutine, Iterable

from starlette.concurrency import run_in_threadpool

from .types import WSGIApp


def empty_wsgi_app() -> WSGIApp:
    """Return a WSGI app that always returns 404 Not Found."""

    def wsgi(environ, start_response):
        status = '404 Not Found'
        body = b'Not Found'
        headers = [('Content-Type', 'text/plain')]
        start_response(status, headers)
        return [body]

    return wsgi


async def call_async(func: Callable, *args, sync=False, **kwargs) -> Coroutine:
    """Call a function in an async manner.

    If the function is synchronous (or the `sync` hint flag is set),
    it is run in the asyncio thread pool.
    """
    if sync or not asyncio.iscoroutinefunction(func):
        return await run_in_threadpool(func, *args, **kwargs)
    return await func(*args, **kwargs)


async def call_all_async(funcs: Iterable[Callable], *args, **kwargs):
    """Call functions in an async manner (with the same arguments)."""
    for func in funcs:
        await call_async(func, *args, **kwargs)
